static char* version = "midi2txt v1.14 by Gnter Nagler (" __DATE__ ")";

#include "midiio.hpp"
#include "miditime.hpp"
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#ifdef __MSDOS__
#include <dos.h>
#else
#endif
#ifdef ADVERTISEMENT
#include "gnadvert.h"
#endif

// values for info_
#define ONLY_ERRORS -2
#define ONLY_LYRICS -1
#define ONLY_CHUNK   0
#define ONLY_SHORT   1
#define ONLY_INFO    2
#define ALL_EVENTS   3

static const char* versioninfo(int version)
{
  switch(version)
  {
  case 0: return "single multichanneltrack";
  case 1: return "several tracks with seperated channels to play all at once";
  case 2: return "several multichanneltracks to play one by one";
  default: return "unknown version";
  }
}

class MidiPrint : public MidiRead
{
public:
  MidiPrint(const char* name);
  virtual ~MidiPrint();

  FILE* outputf_;

  int info_;
  int printunits_;
  int printms_;
  int marktact_;
  long errorcnt_;
  int printdecimal_; // no program, notes names etc.
  int printchannel_;

protected:
  virtual void head(unsigned version, unsigned tracks, unsigned unitperbeat);
  void garbage(FILE* f, long datapos, long datalen);
  virtual void track(int trackno, long length, int channel);
  virtual void endtrack(int trackno);

  virtual void seqnumber(unsigned int seqno);
  virtual void text(int what, unsigned len, char* whattext, unsigned char* txt);
  virtual void meta(int what, unsigned len, unsigned char* data);
  virtual void meta(int what, FILE* f, long datapos, long datalen);
  virtual void end();
  virtual void prefixchannel(unsigned char channel);
  virtual void prefixport(unsigned char port);
  virtual void smpteofs(int mode, int hour, int min, int sec, int frame, int fracframe);
  virtual void tact(int nom, int denom, int unitsperbeat, int notes32perbeat);
  virtual void tempo(unsigned long ticks);
  virtual void key(int signature, int isminor);

  virtual void program(int prg, int channel);
  virtual void control(int channel, int what, int value);
    virtual void highbank(int channel, int val);
    virtual void wheel(int channel, int val);
    virtual void breath(int channel, int val);
    virtual void foot(int channel, int val);
    virtual void portamentotime(int channel, int val);
    virtual void data(int channel, int val);
    virtual void volume(int channel, int val);
    virtual void balance(int channel, int val);
    virtual void expression(int channel, int val);
    virtual void lowbank(int channel, int val);
    virtual void hold(int channel, int val);
    virtual void reverb(int channel, int val);
    virtual void chorus(int channel, int val);
    virtual void datainc(int channel, int val);
    virtual void datadec(int channel, int val);
    virtual void lowrpn(int channel, int val);
    virtual void highrpn(int channel, int val);
    virtual void pitchbendrange(int channel, int val);
  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void time(unsigned long ticks);
  virtual void pitchbend(int channel, int val);
  virtual void polyaftertouch(int channel, int note, int val);
  virtual void aftertouch(int channel, int val);
  virtual void songpos(unsigned pos);
  virtual void songselect(unsigned char song);
  virtual void tunerequest();
  virtual void timingclock();
  virtual void start();
  virtual void cont();
  virtual void stop();
  virtual void activesense();
  virtual void sysex(unsigned syslen, unsigned char* sysdata);
  virtual void sysex(FILE* f, long sysdatapos, long sysdatalen);
  virtual void gmreset();
  virtual void gm2reset();
  virtual void gsreset();
  virtual void gsexit();
  virtual void xgreset();
  virtual void endmidi();
  virtual void error(const char* msg);

  virtual void percent(int perc);

  void printchannel(int channel);
  void printchar(int c);
  void printcurpos(unsigned long units);

private:
  int indent_;
  MidiTimeTable* tactinfo_;

  char hastempochanges_; // can't print milliseconds in tracks 2,3... if tempo changes

  int nl_;
  int endmark_;  // each track should contain FF 2F 00 as last command

  void indent();
  void hex(FILE* f, long pos, long len, FILE* outputf);
  void printxf();
  int runxftrack(int trackno);

};

MidiPrint::MidiPrint(const char* name) : MidiRead(name, NULL, MIDI_PRELOAD)
{
  outputf_ = 0;
  indent_ = 0;
  info_ = ALL_EVENTS;
  marktact_ = printms_ = 0;
  nl_ = 1;
  errorcnt_ = 0;
  hastempochanges_ = 0;
  printdecimal_ = 0;
  printchannel_ = 0;
  tactinfo_ = 0;
}

MidiPrint::~MidiPrint()
{
  delete tactinfo_;
  tactinfo_ = 0;
}

void MidiPrint::indent()
{
  for (; indent_ > 0; indent_--)
    fputc(' ', outputf_);
}

void MidiPrint::head(unsigned version, unsigned tracks, unsigned unitperbeat)
{
  if (marktact_)
  {
    tactinfo_ = new MidiTimeTable(midiname(), getf());
    if (tactinfo_)
    {
      if (!tactinfo_->run())
      {
        delete tactinfo_;
        tactinfo_ = 0;
      }
    }
  }

  if (midiname_)
    fprintf(outputf_, "// %s\n", midiname_);
  if (getcurpos() != 14)
    fprintf(outputf_, " // Warning: %ld bytes garbage at beginning of file\n", getcurpos() - 14);
  if (tracks_ == 0 && info_ == ONLY_ERRORS)
    error("no track found");
  if (info_ == ONLY_SHORT || info_ == ONLY_LYRICS || info_ == ONLY_ERRORS)
    return;
  fprintf(outputf_, "mthd\n");
  fprintf(outputf_, "  version %d // %s\n", version, versioninfo(version));
  fprintf(outputf_, "  // %d track%s\n", tracks, tracks != 1 ? "s" : "");
  fprintf(outputf_, "  unit %d // is 1/4\n", unitperbeat);
  if (tracklen_ > 6)
  {
    fprintf(outputf_, "// Warning: %ld bytes garbage in header:\n", tracklen_ - 6);
    hex(getf(), getcurpos(), tracklen_ - 6, outputf_);
  }
  fprintf(outputf_, "end mthd\n");
}

void MidiPrint::track(int trackno, long /*length*/, int channel)
{
  endmark_ = 0;
  if (info_ == ONLY_SHORT || info_ == ONLY_LYRICS || info_ == ONLY_ERRORS)
    return;
  fprintf(outputf_, "\nmtrk");
  if (channel >= 0)
    fprintf(outputf_, "(%d)", channel+1);
  else if (channel == MULTICHANNEL)
    fprintf(outputf_, " // multichannel track");
  fprintf(outputf_, "  // track %d", trackno);

  fprintf(outputf_, "\n");
}

void MidiPrint::endtrack(int /*trackno*/)
{
  if (!endmark_ && info_ != ONLY_CHUNK)
    fprintf(outputf_, " // Warning: end of track meta event missing\n");

  if (info_ == ONLY_SHORT || info_ == ONLY_ERRORS)
    return;
  if (info_ == ONLY_LYRICS)
  {
    if (!nl_)
    {
      nl_ = 1;
      fprintf(outputf_, "\n");
    }
    return;
  }
  if (info_ >= ALL_EVENTS)
    fprintf(outputf_, "\n");
  fprintf(outputf_, "end mtrk\n");
}

void MidiPrint::printchar(int c)
{
  if (c < 0)
    return;
  if (c == '\\')
    fprintf(outputf_, "\\\\");
  else if (c == '"')
    fprintf(outputf_, "\\\"");
  else if (c != 0 && (isprint(c) || strchr("", c) != 0))
    fputc(c, outputf_);
  else
    fprintf(outputf_, "\\x%02x", c);
}

void MidiPrint::seqnumber(unsigned int seqno)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  fprintf(outputf_, "seqnumber %u\n", seqno);
}

void MidiPrint::text(int what, unsigned len, char* whattext, unsigned char* txt)
{
  if (info_ == ONLY_ERRORS)
    return;
  if (info_ == ONLY_LYRICS)
  {
    if (!txt)
      return;
    if (what == meta_lyric || what == meta_text)
    {
    char* bs;

      while ((bs = (char*)memchr(txt, '\\', len)) != 0)
      *bs = '\n';
      while ((bs = (char*)memchr(txt, '/', len)) != 0)
      *bs = '\n';
      if (*txt == '@')
      {
        txt++; len--;
        switch(*txt)
        {
        case 'K': fprintf(outputf_, "\nKind: "); break;
        case 'V': fprintf(outputf_, "\nVersion: "); break;
        case 'I': fprintf(outputf_, "\nInformation: "); break;
        case 'T': fprintf(outputf_, "\nTitle: "); break;
        case 'L': fprintf(outputf_, "\nLanguage: "); break;
        default:  fprintf(outputf_, "\n@%c: ", *txt); break;
        }
        if (*txt)
        {
          txt++;
          len--;
        }
      }
      if (txt[len-1] == '\r')
      txt[len-1] = '\n';
      nl_ = txt[len-1] == '\n';
      fprintf(outputf_, "%.*s", len, txt);
    }
    return;
  }
  indent();
  if (whattext)
    fprintf(outputf_, "%s ", whattext);
  else
    fprintf(outputf_, "metaevent %d ", what);

  fputc('"', outputf_);
  if (txt)
  {
    while (len-- > 0)
      printchar(*txt++);
  }
  fputc('"', outputf_);
  if (!whattext)
    fprintf(outputf_, " end metaevent");
  fprintf(outputf_, "\n");
}

void MidiPrint::meta(int what, unsigned len, unsigned char* data)
{
unsigned i;

  if (info_ < ALL_EVENTS)
    return;
  indent();
  fprintf(outputf_, "metaevent %d ", what);
  i = 0;
  while (len-- > 0)
  {
    indent_ = 4;
    if (printdecimal_)
      fprintf(outputf_, " %d", *data++);
    else
      fprintf(outputf_, " $%02X", *data++);
    if ( (i % 16) == 15)
    {
      fprintf(outputf_, "\n");
      indent();
    }
    i++;
  }
  fprintf(outputf_, " end metaevent\n");
}

void MidiPrint::meta(int what, FILE* f, long datapos, long datalen)
{
int i;
long oldpos = ftell(f);

  if (info_ < ALL_EVENTS)
    return;
  indent();
  fprintf(outputf_, "metaevent %d ", what);
  i = 0;
  fseek(f, datapos, SEEK_SET);
  while (datalen-- > 0)
  {
    indent_ = 4;
    int c = fgetc(f);
    if (c < 0)
      break;
    if (printdecimal_)
      fprintf(outputf_, " %d", c);
    else
      fprintf(outputf_, " $%02X", c);
    if ( (i % 16) == 15)
    {
      fprintf(outputf_, "\n");
      indent();
    }
    i++;
  }
  fprintf(outputf_, " end metaevent\n");
  fseek(f, oldpos, SEEK_SET);
}

void MidiPrint::end()
{
  endmark_ = 1;
}

void MidiPrint::prefixchannel(unsigned char channel)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  fprintf(outputf_, "prefixchannel %d\n", channel+1);
}

void MidiPrint::prefixport(unsigned char port)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  fprintf(outputf_, "prefixport %d\n", port);
}


void MidiPrint::smpteofs(int mode, int hour, int min, int sec, int frame, int fracframe)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  fprintf(outputf_, "smpteofs %d %02d:%02d:%02d:%d:%d\n",
    mode, hour, min, sec, frame, fracframe);
}

void MidiPrint::tact(int nom, int denom, int unitsperbeat, int notes32perbeat)
{
  if (info_ < ONLY_SHORT)
    return;
  indent();
  fprintf(outputf_, "tact %d / %d %d %d\n", nom, denom, unitsperbeat, notes32perbeat);
  if (!marktact_)
    return;
}

void MidiPrint::tempo(unsigned long ticks)
{
  if (trackno_ != 1 || getcurunit() != 0)
    hastempochanges_ = 1;

  if (info_ < ONLY_SHORT)
    return;
  indent();
  if (ticks != 0)
  {
      fprintf(outputf_, "beats %2.5f", ((float)(60000000.0 / (float)ticks)));
      fprintf(outputf_, " /* %ld microsec/beat */\n",  ticks);
      return;
  }
  else
    fprintf(outputf_, "tempo %ld\n", ticks);
}

void MidiPrint::key(int signature, int isminor)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  fprintf(outputf_, "key \"");
  if (signature != 0)
  {
    fprintf(outputf_, "%d%c ", (signature < 0) ? -signature : signature,
    (signature < 0) ? 'b' : '#');
  }
  else
    fprintf(outputf_, "C");
  fprintf(outputf_, "%s\"\n",  isminor ? "min" : "maj");
}

void MidiPrint::printchannel(int channel)
{
  if (info_ < ALL_EVENTS)
    return;
  if (printchannel_ || getchannel() != channel)
  {
    indent();
    if (channel >= 10 && !printdecimal_)
      fprintf(outputf_, "[%d]", channel+1);
    else if (channel >= 0)
      fprintf(outputf_, "[%d]", channel+1);
    setchannel(channel);
  }
}

void MidiPrint::printcurpos(unsigned long units)
{
  if (info_ < ALL_EVENTS)
    return;
  if (printunits_)
    fprintf(outputf_, " /* U%ld */ ", getcurunit()+units);

  // cant print milliseconds in other than first track if tempo changes
  // because tempo information is only available in first track
  // current tempo microsecperbeat_ initalized to last tempo used in previous track
  // so if tempo does not change it is the correct one for the other tracks too
  if (printms_ && (trackno_ == 1 || !hastempochanges_))
  {
    unsigned long ms, msrest;
    calctime(units, ms, msrest);

    fprintf(outputf_, " /* %ldms */ ", ms);
  }
}

void MidiPrint::time(unsigned long ticks)
{
unsigned long n;
unsigned long u = curunit_;

//  fprintf(stderr, "%lu:%02d\r", getcurmillisec()/ 60000L, int((getcurmillisec() / 1000) % 60));
  indent_ = 2;
  if (ticks == 0)
    printcurpos(0);
  while (ticks > 0)
  {
    unsigned long nextu = tactinfo_ ? tactinfo_->nextMeasure(u) : 0;
    if (!marktact_ || printdecimal_ || !tactinfo_)
      n = ticks;
    else
    {
      n = nextu - u;
      if (n > ticks)
        n = ticks;
    }
    if (info_ < ALL_EVENTS)
      return;
    if (n > 0)
    {
      indent();
      u += n;
      if ((n % unitsperbeat_) == 0 && !printdecimal_)
        fprintf(outputf_, "%ld/4", n / unitsperbeat_);
      else if (unitsperbeat_ % 2 == 0 && (n % (unitsperbeat_ / 2)) == 0 && !printdecimal_)
        fprintf(outputf_, "%ld/8", n / (unitsperbeat_/2));
      else if (unitsperbeat_ % 4 == 0 && (n % (unitsperbeat_ / 4)) == 0 && !printdecimal_)
        fprintf(outputf_, "%ld/16", n / (unitsperbeat_/4));
      else
        fprintf(outputf_, "%ld", n);
      if (!marktact_ || !tactinfo_ || nextu != u)
        fputc(';', outputf_);
      else
      {
        fprintf(outputf_, ".");
        printcurpos(u-curunit_);
        fprintf(outputf_, " // tact %lu\n", (long)tactinfo_->MeasureIndex(u));
        indent_ = 2;
      }
      ticks -= n;
    }
    if (indent_ == 0)
      printcurpos(n);
  }
}

void MidiPrint::program(int channel, int prg)
{
  if (info_ < ONLY_SHORT)
    return;
  indent();
  printchannel(channel);
  if (printdecimal_)
    fprintf(outputf_, "program %d\n", prg);
  else
    fprintf(outputf_, "program %s\n", progname(prg, channel));
}

void MidiPrint::control(int channel, int what, int value)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  if (printdecimal_)
    fprintf(outputf_, "control %d %d", what, value);
  else
    fprintf(outputf_, "control $%02X %d", what, value);
  switch(what)
  {
  case 74: fprintf(outputf_, " // XG brightness"); break;
  case 84: fprintf(outputf_, " // XG portamento"); break;
  case 94: fprintf(outputf_, " // XG effect4"); break;
  }
  fprintf(outputf_, "\n");
}

void MidiPrint::balance(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "balance ");
  if (val == 0 && !printdecimal_)
    fprintf(outputf_, " left\n");
  else if (val == 127 && !printdecimal_)
    fprintf(outputf_, " right\n");
  else
    fprintf(outputf_, " %d\n", val);
}

void MidiPrint::highbank(int channel, int val)
{
  if (info_ < ONLY_INFO)
    return;
  indent();
  printchannel(channel);
  if (printdecimal_)
    fprintf(outputf_, "hbank %d\n", val);
  else
    fprintf(outputf_, "hbank $%02X\n", val);
}

void MidiPrint::wheel(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "wheel %d\n", val);
}

void MidiPrint::breath(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "breath %d\n", val);
}

void MidiPrint::foot(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "foot %d\n", val);
}

void MidiPrint::portamentotime(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "portamentotime %d\n", val);
}

void MidiPrint::data(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "data %d\n", val);
}

void MidiPrint::volume(int channel, int val)
{
  if (info_ < ONLY_INFO)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "volume %d\n", val);
}

void MidiPrint::expression(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "expression %d\n", val);
}

void MidiPrint::lowbank(int channel, int val)
{
  if (info_ < ONLY_INFO)
    return;
  indent();
  printchannel(channel);
  if (printdecimal_)
    fprintf(outputf_, "lbank %d\n", val);
  else
    fprintf(outputf_, "lbank $%02X\n", val);
}

void MidiPrint::hold(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "hold ");
  if (val == 0 && !printdecimal_)
    fprintf(outputf_, " off\n");
  else if (val == 0x7f && !printdecimal_)
    fprintf(outputf_, " on\n");
  else
    fprintf(outputf_, " %d\n", val);
}

void MidiPrint::reverb(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "reverb %d\n", val);
}

void MidiPrint::chorus(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "chorus %d\n", val);
}

void MidiPrint::datainc(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "datainc %d\n", val);
}

void MidiPrint::datadec(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "datadec %d\n", val);
}

void MidiPrint::lowrpn(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "lowrpn %d\n", val);
}

void MidiPrint::highrpn(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "highrpn %d\n", val);
}

void MidiPrint::pitchbendrange(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "pitchbendrange %d\n", val);
}

void MidiPrint::noteon(int channel, int note, int vel)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  if (printdecimal_)
    fprintf(outputf_, "+%d %d;\n", note, vel);
  else
    fprintf(outputf_, "+%s $%02X;\n", notename(note), vel);
}

void MidiPrint::noteoff(int channel, int note, int vel)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  if (printdecimal_)
    fprintf(outputf_, "-%d %d;\n", note, vel);
  else
    fprintf(outputf_, "-%s $%02X;\n", notename(note), vel);
}

void MidiPrint::pitchbend(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "pitchbend %d\n", val);
}

void MidiPrint::polyaftertouch(int channel, int note, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  if (printdecimal_)
    fprintf(outputf_, "polyaftertouch %d %d\n", note, val);
  else
    fprintf(outputf_, "polyaftertouch %s %d\n", notename(note), val);
}

void MidiPrint::aftertouch(int channel, int val)
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  printchannel(channel);
  fprintf(outputf_, "aftertouch %d\n", val);
}

void MidiPrint::songpos(unsigned pos)
{
  if (info_ < ONLY_SHORT)
    return;
  indent();
  fprintf(outputf_, "songpos %d\n", pos);
}

void MidiPrint::songselect(unsigned char song)
{
  if (info_ < ONLY_SHORT)
    return;
  indent();
  fprintf(outputf_, "songselect %d\n", song);
}

void MidiPrint::tunerequest()
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  fprintf(outputf_, "tunerequest\n");
}

void MidiPrint::timingclock()
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  fprintf(outputf_, "timingclock\n");
}

void MidiPrint::start()
{
  if (info_ < ONLY_SHORT)
    return;
  indent();
  fprintf(outputf_, "start\n");
}

void MidiPrint::cont()
{
  if (info_ < ONLY_SHORT)
    return;
  indent();
  fprintf(outputf_, "continue\n");
}

void MidiPrint::stop()
{
  if (info_ < ONLY_SHORT)
    return;
  indent();
  fprintf(outputf_, "stop\n");
}

void MidiPrint::activesense()
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  fprintf(outputf_, "activesensing\n");
}

void MidiPrint::sysex(unsigned syslen, unsigned char* sysdata)
{
unsigned i = 0;

  if (info_ == ONLY_LYRICS)
  {
    if (syslen >= 8 && sysdata[0] == 0 && sysdata[1] == 0x20 && sysdata[2] == 0x24 && sysdata[3] == 0)
    {
      int texttype = sysdata[4];
      unsigned textlen = syslen - 6;
      fprintf(outputf_, "%d:%.*s", texttype,textlen, (const char*)(sysdata+5));
      return;
    }
  }
  if (info_ < ALL_EVENTS)
    return;
  indent();
  syslen--; // last byte is end of sysex (F7)
  fprintf(outputf_, "sysevent ");

  int printascii = 0;
  if (syslen > 6 && memcmp(sysdata, "\x00\x20\x24\x00", 4) == 0 && strchr("\x01\x04\x05\x06\x07", *sysdata) != 0)
  {
    sysdata+=4;
    syslen -= 4;
    if (printdecimal_)
      fprintf(outputf_, "0 32 36 0 %u ", *sysdata);
    else
      fprintf(outputf_, "$00 $20 $24 $00 $%02X ", *sysdata);
    sysdata++; syslen--;
    printascii = 1;
  }
/* // m-live
F0 len 00 20 24 00 01 "..." 0A F7 => chord
F0 len 00 20 24 00 04 "..." 0A F7
F0 len 00 20 24 00 05 "..." 0A F7
F0 len 00 20 24 00 06 "..." 0A F7
F0 len 00 20 24 00 07 "..." 0A F7
*/
  while (syslen > 0)
  {
    indent_ = 4;
    if (printdecimal_)
      fprintf(outputf_, "%d ", *sysdata++);
    else
    {
      if (printascii && isprint(*sysdata))
      {
        fputc('"', outputf_);
        while (syslen > 0 && isprint(*sysdata))
        {
          if (*sysdata == '"')
            fputc('\\', outputf_);
          fputc(*sysdata++, outputf_);
          syslen--;
        }
        fputc('"', outputf_);
        fputc(' ', outputf_);
      }
      else
      {
        fprintf(outputf_, "$%02X ", *sysdata);
        sysdata++; syslen--;
      }
    }
    if ( (i % 16) == 15)
    {
      fprintf(outputf_, "\n");
      indent();
    }
    i++;
  }
  fprintf(outputf_, "end sysevent\n");
}

void MidiPrint::sysex(FILE* f, long sysdatapos, long sysdatalen)
{
int i = 0;
long oldpos = ftell(f);

  if (info_ < ALL_EVENTS)
    return;
  indent();
  sysdatalen--; // last byte is end of sysex (F7)
  fprintf(outputf_, "sysevent ");
  fseek(f, sysdatapos, SEEK_SET);
  while (sysdatalen-- > 0)
  {
    indent_ = 4;
    if (printdecimal_)
      fprintf(outputf_, "$%d ", fgetc(f));
    else
      fprintf(outputf_, "$%02X ", fgetc(f));
    if ( (i % 16) == 15)
    {
      fprintf(outputf_, "\n");
      indent();
    }
    i++;
  }
  fprintf(outputf_, "end sysevent\n");
  fseek(f, oldpos, SEEK_SET);
}

void MidiPrint::gmreset()
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  fprintf(outputf_, "GMReset\n");
}

void MidiPrint::gm2reset()
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  fprintf(outputf_, "GM2Reset\n");
}

void MidiPrint::gsreset()
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  fprintf(outputf_, "GSReset\n");
}

void MidiPrint::xgreset()
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  fprintf(outputf_, "XGReset\n");
}

void MidiPrint::gsexit()
{
  if (info_ < ALL_EVENTS)
    return;
  indent();
  fprintf(outputf_, "GSExit\n");
}

void MidiPrint::hex(FILE* f, long pos, long len, FILE* outputf_)
{
  fseek(f, pos, SEEK_SET);
  unsigned char c[16];
  while (len > 0)
  {
  int n = 16;
  int i;

    if (len < n)
      n = (int)len;
    if (fread(c, n, 1, f) != 1)
      break;
    fprintf(outputf_, "//");
    for (i = 0; i < n; i++)
      fprintf(outputf_, " %02X", c[i]);
    fprintf(outputf_, "\n");
    len -= n;
  }
}

int MidiPrint::runxftrack(int /*trackno*/)
{
  unsigned long trackend;
  trackpos_ = curpos_;

  if (exit_)
    return 0;

  skiptrack_ = 0;
  curunit_ = 0;
  curms_ = currest_ = 0;
  lastcode_ = -1;
  pos_ = curpos_;
  curchannel_ = scanchannel(tracklen_);
  pos_ = trackpos_ = curpos_;
  trackend = trackpos_ + tracklen_;
  lastcode_ = -1;
  if ((options_ & OPTION_NOEVENTS) == 0)
  while (!skiptrack_ && curpos_ < trackpos_ + tracklen_)
  {
    int newpercent = (int)((curpos_ * 100) / midilength());
    if (newpercent != percent_)
      percent(percent_ = newpercent);

    unsigned long delta = getdelta();
    if ( delta >= NOTREALISTIC_PAUSE )
    {
      unrealisticpause(delta);
      skiptrack_ = 1;
      continue;
    }
    time(delta);
    if (exit_)
      return 0;
    curunit_ += delta;
    // warning: MidiRead time calculation is only valid for format 0 files
    // use MidiSerial to calculate time for any midi format
    calctime(delta, curms_, currest_);

    if (runevent(trackend) <= 0)
    {
      skiptrack_ = 0;
      return 0;
    }
    if (exit_)
      return 0;
  }
  skiptrack_ = 0;
  seek(trackend);
  pos_ = curpos_;
  return 1;
}

void MidiPrint::printxf()
{
  fprintf(outputf_, "/*\n");
  fprintf(outputf_, "XFIH\n");
  tracklen_ = getlong();
  runxftrack(0);
  fprintf(outputf_, "end XFIH\n");

  while (curpos_ + 4 < midilength())
  {
    long pos = curpos_;
    if (getlong() != XFKM)
    {
      seek(pos);
      break;
    }
    fprintf(outputf_, "XFKM\n");
    tracklen_ = getlong();
    runxftrack(0);
    fprintf(outputf_, "end XFKM\n");
  }
  fprintf(outputf_, "*/\n");
}

void MidiPrint::endmidi()
{
  if (info_ < ALL_EVENTS && info_ != ONLY_ERRORS)
    return;
  if (getpos() < midilength())
  {
  long len = midilength() - getpos();

    if (len >= 16)
    {
      long endpos = getpos();
      if (getlong() == XFIH)
      {
        printxf();
        endpos = getpos();
      }
      seek(endpos);
      if (endpos >= midilength())
        return;
    }
    fprintf(outputf_, " // Warning: %ld bytes garbage at end of file\n", len);
    if (info_ != ONLY_ERRORS)
    {
      hex(getf(), getpos(), len, outputf_);
    }
  }
}

void MidiPrint::percent(int perc)
{
  if (info_ >= ONLY_CHUNK)
    fprintf(stderr, "%-3d%%\r", perc);
}

void MidiPrint::error(const char* msg)
{
  fprintf(outputf_, "// error: %s\n", msg);
  errorcnt_++;
}

static int usage()
{
  PRINTADVERTDOS();
  printf("usage: midi2txt [-error][-chunk][-short][-info][-lyric][-tact][-units][-ms][-dec][-channel] file.mid [file.txt]\n");
  printf("options:\n");
  printf("-error\tchecks file only. Produces only error report\n");
  printf("-chunk\tprints structure of midi file (header, tracks)\n");
  printf("-short\tprints short info about text, tracknames, programs\n");
  printf("-info\tprints info about text, tracknames, programs, controls\n");
  printf("-lyric\tprints lyrics only\n");
  printf("-tact\tmarks end of tact bars with .\n");
  printf("-units\tprints position of all events in midi units.\n");
  printf("-ms\tprints position of all events in milliseconds (1st track only).\n");
  printf("-dec\tprint decimal values instead of notes, channels, programs,pauses...\n");
  printf("-channel\tprint channel before every event\n");
  return 1;
}

#ifdef NOMAIN
int main_midi2txt(int argc, char**argv)
#else
int main(int argc, char**argv)
#endif
{
int info = ALL_EVENTS;
int tact = 0, printunits = 0, printms = 0, ret = 0, printdec = 0, printchannel = 0;
char* midiname = 0;

  argc--; argv++;

  while (argc > 0 && **argv == '-')
  {
    if (strncmp(*argv, "-version", 2) == 0)
    {
      printf("%s\n", version);
      argc--; argv++;
      if (argc == 0)
    return 0;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-error", 2) == 0)
    {
      info = ONLY_ERRORS; // check if errors in midi
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-decimal", 2) == 0)
    {
      printdec = 1;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-chunk", 4) == 0)
    {
      info = ONLY_CHUNK;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-channel", 4) == 0)
    {
      printchannel = 1;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-short", 2) == 0)
    {
      info = ONLY_SHORT;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-info", 2) == 0)
    {
      info = ONLY_INFO;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-lyric", 2) == 0)
    {
      info = ONLY_LYRICS;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-tact", 2) == 0)
    {
      tact = 1;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-ms", 2) == 0)
    {
      printms = 1;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-units", 2) == 0)
    {
      printunits = 1;
      argc--; argv++;
      continue;
    }
    printf("invalid option %s\n", *argv);
    argc--; argv++;
    return usage();
  }
  if (argc == 0)
    return usage();

  midiname = *argv++; argc--;
  MidiPrint midi(midiname);
  if (!midi.getf())
  {
    perror(midiname);
    return 1;
  }
  if (argc > 0)
  {
    midi.outputf_ = fopen(*argv, "w");
    if (!midi.outputf_)
    {
      perror(*argv);
      return 1;
    }
  }
  else
    midi.outputf_ = stdout;
  fprintf(stderr, "%s\n", midiname);
  midi.info_ = info;
  midi.marktact_ = tact;
  midi.printunits_ = printunits;
  midi.printms_ = printms;
  midi.printdecimal_ = printdec;
  midi.printchannel_ = printchannel;
  if (info == ONLY_CHUNK)
    midi.options_ = OPTION_NOEVENTS;
  if (!midi.run())
  {
    fprintf(midi.outputf_, "\n%s: midi error at 0x%04lx\n", midiname, midi.getpos());
    ret = 1;
  }
  else
    ret = 0;
  if (midi.outputf_ && midi.outputf_ != stdout)
    fclose(midi.outputf_);
  return ret;
}

