#ifndef __MIDICOLLECT_HPP
#define __MIDICOLLECT_HPP

#include "midiio.hpp"

class EventInfo
{
public:
  EventInfo();
  virtual ~EventInfo();

  virtual const int EventType() const { return 0; }
  virtual const char* EventName() { return 0; }
  virtual int ready() { return 1; }
  virtual int started() { return 1; }
  virtual int finished() { return 1; }
  virtual unsigned long endunit() const { return startunit_; }
  virtual unsigned long endms() const { return startms_; }

  int track_;                   // 1-255, 0 if event is in midi header
  unsigned char channel_;       // 1-16, 0 if event has no channel
  unsigned long startunit_;
  unsigned long startms_;
};

class NoteInfo : public EventInfo
{
public:
  NoteInfo();

  virtual const int EventType() const { return 'N'; }
  virtual const char* EventName() { return "Note"; }
  virtual int ready() { return started() && finished(); }
  virtual int started() { return started_; }
  virtual int finished() { return finished_; }
  virtual unsigned long endunit() const { return endunit_; }
  virtual unsigned long endms() const { return endms_; }
  virtual unsigned long length() { return endunit_ - startunit_; }
  virtual unsigned long duration() { return endms_ - startms_; }

  unsigned long endunit_;
  unsigned long endms_;
  unsigned char note_;
  unsigned char vel_, offvel_;
  char started_, finished_;
};

class ProgramInfo : public EventInfo
{
public:
  ProgramInfo();

  virtual const int EventType() const { return 'P'; }
  virtual const char* EventName() { return "Program"; }

  unsigned char program_;
};

class ControlInfo : public EventInfo
{
public:
  ControlInfo();

  virtual const int EventType() const { return 'C'; }
  virtual const char* EventName() { return "Control"; }

  unsigned char controlnr_;
  unsigned char controlvalue_;
};

class HeaderInfo : public EventInfo
{
public:
  HeaderInfo(int unitsperbeat, int version, int trackcount);
  HeaderInfo(MidiRead*);

  virtual const int EventType() const { return 'H'; }
  virtual const char* EventName() { return "Header"; }

  int unitsperbeat_, version_, trackcount_;
};

class TempoInfo : public EventInfo
{
public:
  TempoInfo(long microsecperbeat);

  virtual const int EventType() const { return 'T'; }
  virtual const char* EventName() { return "Tempo"; }

  long  microsecperbeat_;

  float beatspermin();
  int int_beatspermin();
};

class TactInfo : public EventInfo
{
public:
  TactInfo(int nomin, int denom, int unitsperbeat = 24, int notes32perbeat = 8);
  virtual const int EventType() const { return 'B'; }
  virtual const char* EventName() { return "Bar"; }

  int nomin_, denom_; // e.g. 3/4
  int unitsperbeat_, notes32perbeat_; // e.g. 24 8
};

class TextInfo : public EventInfo
{
public:
  TextInfo(const char* text, unsigned textlen, int texttype);
 ~TextInfo();

  virtual const int EventType() const { return 't'; }
  virtual const char* EventName() { return "Text"; }

  const char* gettext();
  void settext(const char* text, unsigned textlen);
  int texttype_; // meta values
  unsigned textlen_;// string can contain \0 characters!

protected:
  char* text_;
};

class PitchbendInfo : public EventInfo
{
public:
  PitchbendInfo();

  virtual const int EventType() const { return 'p'; }
  virtual const char* EventName() { return "Pitch"; }

  int pitchbend_; // e.g. -8192..8191
};

class AftertouchInfo : public EventInfo
{
public:
  AftertouchInfo();

  virtual const int EventType() const { return 'a'; }
  virtual const char* EventName() { return "Aftertouch"; }

  unsigned char aftertouch_;
};

class PolyAftertouchInfo : public EventInfo
{
public:
  PolyAftertouchInfo();

  virtual const int EventType() const { return 'A'; }
  virtual const char* EventName() { return "Polyaftertouch"; }

  unsigned char key_;
  unsigned char value_;
};

class KeyInfo : public EventInfo
{
public:
  KeyInfo(int key = 0, int minor = 0);

  virtual const int EventType() const { return 'k'; }
  virtual const char* EventName() { return "Key"; }

  const char* keyname();

  int key_; // -7..+7  (number of <b,>#, 0=C)
  int minor_;
};

class SysexInfo : public EventInfo
{
public:
  SysexInfo(const unsigned char* data, unsigned len);
 ~SysexInfo();

  virtual const int EventType() const { return 'x'; }
  virtual const char* EventName() { return "Sysex"; }

  const unsigned char* getdata();
  unsigned getdatalen();

protected:
  unsigned char* data_;
  unsigned len_;
};

class MetaInfo : public EventInfo
{
public:
  MetaInfo(int what, unsigned len, const unsigned char* data);
 ~MetaInfo();

  virtual const int EventType() const { return 'M'; }
  virtual const char* EventName() { return "Meta"; }

  int what_;
  const unsigned char* getdata();
  unsigned getdatalen();

protected:
  unsigned char* data_;
  unsigned len_;
};


class EventInfoList
{
public:
  EventInfoList();
  virtual ~EventInfoList();

  void clear();
  int find(EventInfo* event); // returns -1 if not found or index
  int add(EventInfo* event);
  int insert(int index, EventInfo* event);
  int remove(EventInfo* event);
  int kill(EventInfo* event);
  void sorttime();

  EventInfo* remove(int index);
  int kill(int index);

  unsigned long lastunit();
  int resolution();

  EventInfo** event_;
  int count_;

private:
  int maxcount_;
  int resize();
};

class MidiCollect : public MidiSerial
{
public:
  MidiCollect(const char* filename, const char* eventmask = NULL, char preload = 0);
  ~MidiCollect();

  char option_delete_events_; // default=1  MidiCollect removes all events after use
  int channel_; // -1 if all, 0 for no channel, 1-16
  int track_; // -1 if all, 0 for header only

  int addevent(EventInfo* event);
  void killevent(EventInfo* event);

  void seteventmask(const char* eventlist);

  virtual void StartEvents();
  virtual void Event(EventInfo* event);
  virtual void UnstartedEvent(EventInfo* event);
  virtual void UnfinishedEvents();
  virtual void EndEvents();

protected:
  virtual void head(unsigned version, unsigned tracks, unsigned unitperbeat);

  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void program(int channel, int prg);
  virtual void control(int channel, int what, int val);
  virtual void pitchbend(int channel, int val);
  virtual void aftertouch(int channel, int val);
  virtual void polyaftertouch(int channel, int key, int val);

  virtual void tempo(unsigned long microsecperbeat);
  virtual void tact(int nom, int denom, int unitsperbeat, int notes32perbeat);

  virtual void text(int what, unsigned len, char* whattext, unsigned char* txt);

  virtual void key(int signature, int isminor); // C=0, -7=7flats +7=7sharps

  virtual void sysex(unsigned len, unsigned char* sysdata);
  virtual void meta(int what, unsigned len, unsigned char* data);

   virtual void prefixchannel(unsigned char channel);
   virtual void prefixport(unsigned char port);
   virtual void seqnumber(unsigned int seqno);
   virtual void smpteofs(int mode, int hour, int min, int sec, int frame, int fracframe);

  int startevent(EventInfo* event);
  void stopnote(NoteInfo* note);

  virtual int ignore(int evtype, int channel, int track);

  virtual void endmidi();

protected:
  const char* eventlist_;
  unsigned long collectunit_;
  long recordcnt_;
  long lastunit_;

  EventInfoList list_;

  int curtact_nomin_, curtact_denom_, curtact_units_;
  long lasttact_unit_;

  int curtactunits();

  int findnote(NoteInfo* note);

  void collect();
  void fireevent(EventInfo* event);

  virtual void outofmemory();
};

#endif
